/*
 * Copyright (c) 2013-2018 Intel Corporation. All rights reserved.
 * Copyright (c) 2016-2018 Cisco Systems, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef _OFI_STR_H_
#define _OFI_STR_H_

#include <string.h>
#include <stdint.h>

#include <rdma/fabric.h>
#include "config.h"

#ifdef __cplusplus
extern "C" {
#endif

int ofi_rm_substr(char *str, const char *substr);
int ofi_rm_substr_delim(char *str, const char *substr, const char delim);
char **ofi_split_and_alloc(const char *s, const char *delim, size_t *count);
void ofi_free_string_array(char **s);
char *ofi_tostr_size(char *str, size_t len, uint64_t size);
char *ofi_tostr_count(char *str, size_t len, uint64_t count);
void ofi_strncatf(char *dest, size_t n, const char *fmt, ...);
int ofi_nic_tostr(const struct fid *fid_nic, char *buf, size_t len);

#ifdef __cplusplus
}
#endif

#endif /* _OFI_STR_H_ */
