/*
 * Copyright (C) 2025 Cornelis Networks.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#ifndef _FI_PROV_OPX_HFISVC_H_
#define _FI_PROV_OPX_HFISVC_H_

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#ifndef OPX_HFISVC_ENABLED_DEFAULT
#define OPX_HFISVC_ENABLED_DEFAULT (1)
#endif

extern int opx_hfisvc_log_enabled;

#ifdef OPX_HFISVC_DEBUG
#define OPX_HFISVC_DEBUG_LOG(fmt, ...)                                                                  \
	do {                                                                                            \
		if (opx_hfisvc_log_enabled) {                                                           \
			fprintf(stderr, "(%d) %s:%s():%d " fmt, getpid(), __FILE__, __func__, __LINE__, \
				##__VA_ARGS__);                                                         \
		}                                                                                       \
	} while (0)
#else
#define OPX_HFISVC_DEBUG_LOG(fmt, ...)
#endif

#endif
